/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "fobprivate.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

int fob=-1;
#define NUMREC 4



/* Write a buffer to the FOB */

void fobWrite(unsigned char *buf)
{
  int rv;

  /* Write the buffer and check its all gone */
  rv= write(fob, buf, strlen((char *) buf));
  if (rv!=strlen((char *) buf)) printf("rv!=strlen(buf)\n");
  tcflush(fob, TCIFLUSH);
}

void fobWriteN(unsigned char *buf, int n)
{
  int rv;

  /* Write the buffer and check its all gone */
  rv= write(fob, buf, n);
  if (rv!=n) printf("rv!=strlen(buf)\n");
  tcflush(fob, TCIFLUSH);
}



/* Read a given number of characters from the FOB */

void fobRead(int num, unsigned char *buf)
{
  fd_set pf;
  int rv, totread=0;

  do {
    FD_ZERO(&pf);
    FD_SET(fob, &pf);
    select(200, &pf, NULL, NULL, NULL);
    rv= read(fob, &buf[totread], num-totread);
    totread+=rv;
  } while (totread!=num);
  
  buf[num]=0;
}



/* Open a connection with the FOB */

int fobOpen(char *name)
{
  struct termios mysettings;
  unsigned char id[]={79, 15, 0};
  unsigned char conf[]={80, 50, 0, 0};
  unsigned char quat[]={0, 93, 0};
  unsigned char grp[]={80, 35, 1, 0};
  unsigned char buf[1000];
  int i;
  
  /* Open the serial port to the FOB */
  fob = open(name, O_RDWR);

  if (fob == -1) {
    fprintf(stderr, "Could not open FOB on device %s : ", name);
    perror("");
    return(0);
  }

  /* Read the present state of the port */
  if (tcgetattr(fob, &mysettings) == -1) {
    perror("Could not read FOB ioctl settings ");
    return(0);
  }

  /* Update it with our settings */
  mysettings.c_iflag= IGNBRK|IGNPAR;
  mysettings.c_cc[CEOF]= 1;
  mysettings.c_lflag= 0;
  cfsetispeed(&mysettings, B38400);
  cfsetospeed(&mysettings, B38400);

  /* Now set those settings onto the port */
  if (tcsetattr(fob, TCSANOW, &mysettings) == -1) {
    perror("Could not set FOB ioctl settings ");
    return(0);
  }

  /* Ask for system id */  
  fprintf(stderr, "Reading ID... ");
  fobWrite(id);
  fobRead(6, buf);
  fprintf(stderr, "done - its [%s]\n", buf);

  if (strcmp((const char *) buf, "6DERC ")) {
    printf("Unsupported type - only know about 6DERC\n");
    return(0);
  }

  /* Configure the FOB */
  sleep(1);
  fprintf(stderr, "Configuring for %i birds... ", NUMREC);
  conf[2]= NUMREC;
  fobWrite(conf);
  fprintf(stderr, "done\n");
  sleep(1);

  /* Ask for position and quaternion */
  fprintf(stderr, "Asking for position and quaternion... ");
  for (i=0; i<NUMREC-1; i++) {
    quat[0]=240+2+i; /* Address each bird */
    fobWrite(quat);
  }
  fprintf(stderr, "done\n");

  /* Switch to group mode */
  fprintf(stderr, "Switching to group mode... ");
  fobWrite(grp);
  fprintf(stderr, "done\n");

  return 1;
}




float conv(int lo, char hi)
{
  int comb;

  /* Take the 7 LSBs of lo and the 7 LSBs of hi to form a 14 bit number */
  comb= (lo&127) | ((hi&127)<<7);

  /* The MSBs value is -ve */
  if (comb>=8192) comb-=16384;

  /* Scale to give a float [-1:1] */
  return ((float) comb)/8192;
}

/* Endlessly obtain data from the FOB */

void fobLoop(int debug)
{
  unsigned char pt[]={66, 0};
  unsigned char buf[1000];
  int i, lc=0;
  struct timeval st, ed;

  gettimeofday(&st, NULL);

  fprintf(stderr, "All OK\n");

  while (1) {
    lc++;
    
    /* Ask for data (all birds) */
    fobWrite(pt);

    for (i=0; i<NUMREC-1; i++) {

      /* Get reply (for single bird) */
      fobRead(15 , buf);

      /* Check phasing bit */
      if (buf[0]<128)
      {
	printf("Out of sync, buf[0]=%i\n", buf[0]);
	exit(1);
      }
      
      /* Check correct reciver */
      if (buf[14]-2!=i) {
	printf("Reading reciever %i when expecting %i\n", buf[14]-2, i);
	exit(1);
      }

      /* Get x,y,z position */
      shared->x[i]= conv(buf[0], buf[1])*144;
      shared->y[i]= conv(buf[2], buf[3])*144;
      shared->z[i]= conv(buf[4], buf[5])*144;

      /* Get quaternion */
      shared->q0[i]= conv(buf[6], buf[7]);
      shared->q1[i]= conv(buf[8], buf[9]);
      shared->q2[i]= conv(buf[10], buf[11]);
      shared->q3[i]= conv(buf[12], buf[13]);
    }

    /* Print out record if debugging set */
    if (lc==100 && debug) {
      gettimeofday(&ed, NULL);
      printf("time %f\n", lc*1.0/((ed.tv_sec-st.tv_sec)+(ed.tv_usec-st.tv_usec)/1.0E+6));
      printf("pos %f %f %f quat %f %f %f %f\n", shared->x[0], shared->y[0], shared->z[0], shared->q0[0], shared->q1[0], shared->q2[0], shared->q3[0]);
      lc=0;
      st=ed;
    }
  }
}



/* Routine to close the FOB */

void fobClose(void)
{
  printf("Closing FOB connection\n");

  close(fob);
}
