/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "it30private.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>

int intertrax=-1;



/* Write a buffer to the InterTrax - these are just single char commands */

void interTraxWrite(char *buf)
{
  int rv;

  /* Write the buffer and check its all gone */
  rv= write(intertrax, buf, strlen(buf));
  if (rv!=strlen(buf)) printf("rv!=strlen(buf)\n");
}



/* Read for a certain length of time from the InterTrax box */

void interTraxReadTimed(float sec, unsigned char *buf)
{
  fd_set pf;
  struct timeval t;
  int rv, totread=0;

  FD_ZERO(&pf);
  FD_SET(intertrax, &pf);

  /* Set timeout */
  t.tv_sec= sec;
  t.tv_usec= 0;

  /* Read from the InterTrax box until timed out */
  while (select(200, &pf, NULL, NULL, &t)) {
    FD_ZERO(&pf);
    FD_SET(intertrax, &pf);
    rv= read(intertrax, &buf[totread], 1000);
    totread+=rv;
  }
  
  buf[totread]=0;
}



/* Read a given number of characters from the InterTrax box */

void interTraxRead(int num, unsigned char *buf)
{
  fd_set pf;
  int rv, totread=0;

  do {
    FD_ZERO(&pf);
    FD_SET(intertrax, &pf);
    select(200, &pf, NULL, NULL, NULL);
    rv= read(intertrax, &buf[totread], num-totread);
    totread+=rv;
  } while (totread!=num);
  
  buf[num]=0;
}



/* Open a connection with the InterTrax box */

int it30Open(char *name)
{
  unsigned char buf[1000];
  struct termios mysettings;

  /* Open the serial port to the InterTrax box */
  intertrax= open(name, O_RDWR);

  if (intertrax==-1) {
    fprintf(stderr, "Could not open InterTrax on device %s : ", name);
    perror("");
    return 0;
  }

  /* Read the present state of the port */
  if (tcgetattr(intertrax, &mysettings) == -1) {
    perror("Could not read InterTrax ioctl settings ");
    return 0;
  }

  /* Update it with our settings */
  mysettings.c_iflag= IGNBRK|IGNPAR;
  mysettings.c_cc[CEOF]= 1;
  mysettings.c_lflag= 0;

  /* Set baud rate to 1200 */
  cfsetispeed(&mysettings, B1200);
  cfsetospeed(&mysettings, B1200);

  /* Now set those settings onto the port */
  if (tcsetattr(intertrax, TCSANOW, &mysettings) == -1) {
    perror("Could not set InterTrax ioctl settings ");
    return 0;
  }

  /* Spec says we need to toggle RTS line - appears to work without it */
  /* This gets toggled by merely opening the port? */

  /* Flush buffer - read for 1 sec ignoring whats read */
  interTraxReadTimed(1.0, buf);
  
  /* Switch to command mode */
  interTraxWrite("c");
  
  /* Should respond with "o" */
  interTraxReadTimed(1.0, buf);
  if (strcmp((char *) buf, "o")) {
    printf("InterTrax responded with [%s] when expecting a \"o\"\n", buf);
    return 0;
  }

  /* Switch to data mode */
  interTraxWrite("d");

  /* Should respond with "d" */
  interTraxReadTimed(1.0, buf);
  if (strcmp((char *) buf, "d")) {
    printf("InterTrax responded with [%s] when expecting a \"d\"\n", buf);
    return 0;
  }

  /* Set baud rate to 38400 */
  cfsetispeed(&mysettings, B38400);
  cfsetospeed(&mysettings, B38400);

  /* Now set those settings onto the port */
  if (tcsetattr(intertrax, TCSANOW, &mysettings) == -1) {
    perror("Could not set Polhemus ioctl settings ");
    return 0;
  }

  printf("Connected to InterTrax\n");
  return 1;
}



/* Endlessly read the InterTrax box */

void it30Loop(int debug)
{
  unsigned char buf[10];
  struct timeval st, ed;
  int lc=0;

  gettimeofday(&st, NULL);

  while (1) {
    lc++;

    /* Request all 3 angles */
    interTraxWrite("3");

    /* Get results - a 6 byte packet */
    interTraxRead(6, buf);

    /* Store in shared memory */
    shared->yaw[0]= buf[1]*256+buf[0];
    shared->pitch[0]= buf[3]*256+buf[2];
    shared->roll[0]= buf[5]*256+buf[4];

    /* Print out record if debugging set */
    if (lc==100 && debug) {
      gettimeofday(&ed, NULL);
      printf("time %f ", lc*1.0/((ed.tv_sec-st.tv_sec)+(ed.tv_usec-st.tv_usec)/1.0E+6));
      printf("rpy  %i %i %i\n", shared->roll[0], shared->pitch[0], shared->yaw[0]);
      lc=0;
      st=ed;
    }
  }
}



/* Routine to close the InterTrax box */

void it30Close(void)
{
  printf("Closing InterTrax connection\n");

  close(intertrax);
}

