/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "polprivate.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

typedef struct {
  float x,y,z;
} _Vector;

typedef struct {
  float x,y,z,w;
} _Quaternion;

typedef struct {
  float mat[4][4];
} _Matrix;

_Vector vectorSet (float x, float y, float z)
{
  _Vector v;

  v.x= x;
  v.y= y;
  v.z= z;
  return v;
}

_Vector vectorMult (_Vector in, _Matrix matrix)
{
  _Vector out;

  out.x = in.x*matrix.mat[0][0] + in.y*matrix.mat[0][1] + in.z*matrix.mat[0][2] + matrix.mat[0][3];
  out.y = in.x*matrix.mat[1][0] + in.y*matrix.mat[1][1] + in.z*matrix.mat[1][2] + matrix.mat[1][3];
  out.z = in.x*matrix.mat[2][0] + in.y*matrix.mat[2][1] + in.z*matrix.mat[2][2] + matrix.mat[2][3];
  
  return out;
}

_Quaternion quaternionMatrixConvert (_Matrix m)
{
  _Quaternion quaternion;
  float a, b, c, d, norm;
  float i4a, i2b, inorm;

#define M(X,Y) m.mat[Y-1][X-1]

  a = 0.25*(1.0 + M(1,1) + M(2,2) + M(3,3));

  if (a > 0.0) 
  {
    a = sqrt(a);
    i4a=1.0/(4.0*a);
    b = (M(2,3) - M(3,2))*i4a;
    c = (M(3,1) - M(1,3))*i4a;
    d = (M(1,2) - M(2,1))*i4a;
  }
  else
  {
    a = 0.0;
    b = -0.5*(M(2,2) + M(3,3));
      
    if (b > 0.0)
    {
      b = sqrt(b);
      i2b = 1.0/(2.0*b);
      c = M(1,2)*i2b;
      d = M(1,3)*i2b;
    }
    else
    {
      b = 0.0;
      c = 0.5*(1.0 - M(3,3));
	  
      if (c > 0.0) 
      {
	c = sqrt(c);
	d = M(2,3)/(2.0*c);
      }
      else
      {
	c = 0.0;
	d = 1.0;
      }
    }
  }

#undef M

  norm = sqrt(a*a + b*b + c*c + d*d);
  
  inorm=1.0/norm;
  a *= inorm;
  b *= inorm;
  c *= inorm;
  d *= inorm;
  
  quaternion.w = a;
  quaternion.x = b;
  quaternion.y = c;
  quaternion.z = d;

  return quaternion;
}

_Matrix matrixQuaternionConvert (_Quaternion q)
{
  _Matrix m;

  m.mat[0][0]=1.0-2.0*(q.y*q.y+q.z*q.z);
  m.mat[0][1]=2.0*(q.x*q.y-q.w*q.z);
  m.mat[0][2]=2.0*(q.x*q.z+q.w*q.y);
  m.mat[0][3]=0.0;

  m.mat[1][0]=2.0*(q.x*q.y+q.w*q.z);
  m.mat[1][1]=1.0-2.0*(q.x*q.x+q.z*q.z);
  m.mat[1][2]=2.0*(q.y*q.z-q.w*q.x);
  m.mat[1][3]=0.0;

  m.mat[2][0]=2.0*(q.x*q.z-q.w*q.y);
  m.mat[2][1]=2.0*(q.y*q.z+q.w*q.x);
  m.mat[2][2]=1.0-2.0*(q.x*q.x+q.y*q.y);
  m.mat[2][3]=0.0;

  m.mat[3][0]=0.0;
  m.mat[3][1]=0.0;
  m.mat[3][2]=0.0;
  m.mat[3][3]=1.0;

  return m;
}

_Quaternion quaternionMult (_Quaternion q1, _Quaternion q2)
{
  _Quaternion q;

  q.w= q1.w*q2.w-q1.x*q2.x-q1.y*q2.y-q1.z*q2.z;
  q.x= q1.w*q2.x+q1.x*q2.w+q1.y*q2.z+q1.z*q2.y;
  q.y= q1.w*q2.y+q1.y*q2.w+q1.z*q2.x+q1.x*q2.z;
  q.z= q1.w*q2.z+q1.z*q2.w+q1.x*q2.y+q1.y*q2.x;

  return q;
}

_Quaternion getQuaternion (TDM_position pos)
{
  _Matrix m;

  m.mat[0][0]= pos.u[0];
  m.mat[1][0]= pos.u[1];
  m.mat[2][0]= pos.u[2];

  m.mat[0][1]= pos.v[0];
  m.mat[1][1]= pos.v[1];
  m.mat[2][1]= pos.v[2];

  m.mat[0][2]= pos.n[0];
  m.mat[1][2]= pos.n[1];
  m.mat[2][2]= pos.n[2];

  return quaternionMatrixConvert(m);
}

int polhemusLoadDistortionGrid (DistortionGrid *grid, char *filename)
{
  FILE *f;
  int x,y,z,valid,p;

  f= fopen(filename,"r");
  if (f == NULL)
    return 0;

  /* read grid start */
  fscanf(f,"%g",&grid->sx);
  fscanf(f,"%g",&grid->sy);
  fscanf(f,"%g",&grid->sz);

  /* read grid end */
  fscanf(f,"%g",&grid->ex);
  fscanf(f,"%g",&grid->ey);
  fscanf(f,"%g",&grid->ez);

  /* read grid size */
  fscanf(f,"%d",&grid->nx);
  fscanf(f,"%d",&grid->ny);
  fscanf(f,"%d",&grid->nz);

  /* calculate voxel size */
  grid->dx= (grid->ex-grid->sx)/(float)grid->nx;
  grid->dy= (grid->ey-grid->sy)/(float)grid->ny;
  grid->dz= (grid->ez-grid->sz)/(float)grid->nz;

  /* allocate storage */
  grid->valid= calloc(grid->nx*grid->ny*grid->nz,sizeof(unsigned char));
  grid->position= calloc(3*grid->nx*grid->ny*grid->nz,sizeof(float));
  grid->orientation= calloc(4*grid->nx*grid->ny*grid->nz,sizeof(float));

  /* read data */
  for (z= 0; z< grid->nz; z++)
    for (y= 0; y< grid->ny; y++)
      for (x= 0; x< grid->nx; x++)
	{
	  fscanf(f,"%d",&valid);

	  if (valid)
	    {
	      p= z*grid->nx*grid->ny+y*grid->nx+x;

	      grid->valid[p]= 1;

	      fscanf(f,"%f",&grid->position[3*p+0]);
	      fscanf(f,"%f",&grid->position[3*p+1]);
	      fscanf(f,"%f",&grid->position[3*p+2]);

	      fscanf(f,"%f",&grid->orientation[4*p+0]);
	      fscanf(f,"%f",&grid->orientation[4*p+1]);
	      fscanf(f,"%f",&grid->orientation[4*p+2]);
	      fscanf(f,"%f",&grid->orientation[4*p+3]);
	    }
	}

  fclose(f);
  return 1;
}

int polhemusRemoveDistortion (DistortionGrid *grid, int tracker, TDM_position in, TDM_position *out)
{
  int x,y,z;
  float a,b,c;
  float px,py,pz;
  int p,x1,y1,z1;
  _Quaternion dq;
  _Matrix m;
  _Vector v;

  /* map the distorted sensor coordinates back into lab coordinates */
  x= (int)((in.xpos-grid->sx)/grid->dx);
  y= (int)((in.ypos-grid->sy)/grid->dy);
  z= (int)((in.zpos-grid->sz)/grid->dz);

  /* within grid? */
  if (x < 0 || x >= grid->nx-1 || y < 0 || y >= grid->ny-1 || z < 0 || z >= grid->nz-1)
    {
      *out= grid->last[tracker];
      return 0;
    }

  p= z*grid->nx*grid->ny+y*grid->nx+x;
  x1= 1;
  y1= grid->nx;
  z1= grid->nx*grid->ny;

  /* are we surrounded by any invalid grid entries? */
  if (grid->valid[p] == 0 || grid->valid[p+x1] == 0 || grid->valid[p+x1+z1] == 0 || grid->valid[p+z1] == 0 ||
      grid->valid[p+y1] == 0 || grid->valid[p+x1+y1] == 0 || grid->valid[p+x1+y1+z1] == 0 || grid->valid[p+y1+z1] == 0)
    {
      *out= grid->last[tracker];
      return 0;
    }

  /* tri-linearly interpolate to get lab coordinates */
  a= (in.xpos-(grid->sx+x*grid->dx))/grid->dx;
  b= (in.ypos-(grid->sy+y*grid->dy))/grid->dy;
  c= (in.zpos-(grid->sz+z*grid->dz))/grid->dz;

  p= 3*(z*grid->nx*grid->ny+y*grid->nx+x);

  x1= 3;
  y1= 3*grid->nx;
  z1= 3*grid->nx*grid->ny;

  px= (((1.0-a)*(1.0-b)*(1.0-c))*grid->position[p]+
       (a*(1.0-b)*(1.0-c))*grid->position[p+x1]+
       (a*(1.0-b)*c)*grid->position[p+x1+z1]+
       ((1.0-a)*(1.0-b)*c)*grid->position[p+z1]+
       ((1.0-a)*b*(1.0-c))*grid->position[p+y1]+
       (a*b*(1.0-c))*grid->position[p+x1+y1]+
       (a*b*c)*grid->position[p+x1+y1+z1]+
       ((1.0-a)*b*c)*grid->position[p+y1+z1]);

  py= (((1.0-a)*(1.0-b)*(1.0-c))*grid->position[p+1]+
       (a*(1.0-b)*(1.0-c))*grid->position[p+x1+1]+
       (a*(1.0-b)*c)*grid->position[p+x1+z1+1]+
       ((1.0-a)*(1.0-b)*c)*grid->position[p+z1+1]+
       ((1.0-a)*b*(1.0-c))*grid->position[p+y1+1]+
       (a*b*(1.0-c))*grid->position[p+x1+y1+1]+
       (a*b*c)*grid->position[p+x1+y1+z1+1]+
       ((1.0-a)*b*c)*grid->position[p+y1+z1+1]);

  pz= (((1.0-a)*(1.0-b)*(1.0-c))*grid->position[p+2]+
       (a*(1.0-b)*(1.0-c))*grid->position[p+x1+2]+
       (a*(1.0-b)*c)*grid->position[p+x1+z1+2]+
       ((1.0-a)*(1.0-b)*c)*grid->position[p+z1+2]+
       ((1.0-a)*b*(1.0-c))*grid->position[p+y1+2]+
       (a*b*(1.0-c))*grid->position[p+x1+y1+2]+
       (a*b*c)*grid->position[p+x1+y1+z1+2]+
       ((1.0-a)*b*c)*grid->position[p+y1+z1+2]);

  /* store corrected position */
  in.xpos= px;
  in.ypos= py;
  in.zpos= pz;

  /* tri-linearly interpolate to get orientation quaternion */
  p= 4*(z*grid->nx*grid->ny+y*grid->nx+x);

  x1= 4;
  y1= 4*grid->nx;
  z1= 4*grid->nx*grid->ny;

  dq.w= -(((1.0-a)*(1.0-b)*(1.0-c))*grid->orientation[p]+
	  (a*(1.0-b)*(1.0-c))*grid->orientation[p+x1]+
	  (a*(1.0-b)*c)*grid->orientation[p+x1+z1]+
	  ((1.0-a)*(1.0-b)*c)*grid->orientation[p+z1]+
	  ((1.0-a)*b*(1.0-c))*grid->orientation[p+y1]+
	  (a*b*(1.0-c))*grid->orientation[p+x1+y1]+
	  (a*b*c)*grid->orientation[p+x1+y1+z1]+
	  ((1.0-a)*b*c)*grid->orientation[p+y1+z1]);

  dq.x= (((1.0-a)*(1.0-b)*(1.0-c))*grid->orientation[p+1]+
	  (a*(1.0-b)*(1.0-c))*grid->orientation[p+x1+1]+
	  (a*(1.0-b)*c)*grid->orientation[p+x1+z1+1]+
	  ((1.0-a)*(1.0-b)*c)*grid->orientation[p+z1+1]+
	  ((1.0-a)*b*(1.0-c))*grid->orientation[p+y1+1]+
	  (a*b*(1.0-c))*grid->orientation[p+x1+y1+1]+
	  (a*b*c)*grid->orientation[p+x1+y1+z1+1]+
	  ((1.0-a)*b*c)*grid->orientation[p+y1+z1+1]);

  dq.y= (((1.0-a)*(1.0-b)*(1.0-c))*grid->orientation[p+2]+
	  (a*(1.0-b)*(1.0-c))*grid->orientation[p+x1+2]+
	  (a*(1.0-b)*c)*grid->orientation[p+x1+z1+2]+
	  ((1.0-a)*(1.0-b)*c)*grid->orientation[p+z1+2]+
	  ((1.0-a)*b*(1.0-c))*grid->orientation[p+y1+2]+
	  (a*b*(1.0-c))*grid->orientation[p+x1+y1+2]+
	  (a*b*c)*grid->orientation[p+x1+y1+z1+2]+
	  ((1.0-a)*b*c)*grid->orientation[p+y1+z1+2]);

  dq.z= (((1.0-a)*(1.0-b)*(1.0-c))*grid->orientation[p+3]+
	  (a*(1.0-b)*(1.0-c))*grid->orientation[p+x1+3]+
	  (a*(1.0-b)*c)*grid->orientation[p+x1+z1+3]+
	  ((1.0-a)*(1.0-b)*c)*grid->orientation[p+z1+3]+
	  ((1.0-a)*b*(1.0-c))*grid->orientation[p+y1+3]+
	  (a*b*(1.0-c))*grid->orientation[p+x1+y1+3]+
	  (a*b*c)*grid->orientation[p+x1+y1+z1+3]+
	  ((1.0-a)*b*c)*grid->orientation[p+y1+z1+3]);

  /* convert orientation correction to matrix form */
  m= matrixQuaternionConvert(dq);

  /* correct axis vectors */
  v= vectorSet(in.u[0],in.u[1],in.u[2]);
  v= vectorMult(v,m);
  in.u[0]= v.x;
  in.u[1]= v.y;
  in.u[2]= v.z;

  v= vectorSet(in.v[0],in.v[1],in.v[2]);
  v= vectorMult(v,m);
  in.v[0]= v.x;
  in.v[1]= v.y;
  in.v[2]= v.z;

  v= vectorSet(in.n[0],in.n[1],in.n[2]);
  v= vectorMult(v,m);
  in.n[0]= v.x;
  in.n[1]= v.y;
  in.n[2]= v.z;

  /* store this corrected position and orientation, in case we move
     out-of-bounds for the next sample */
  grid->last[tracker]= *out= in;
  return 1;
}
