/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


/*
   This code derived from TDM v1.3 Polhemus driver code
      (C) 1999 Advanced Interfaces Group

   Some parts of code taken from joystick 1.2.14 package
      (C) 1998 David Thompson (dcthomp@mail.utexas.edu)
      (C) 1999 Vojtech Pavlik (vojtech@ucw.cz)

   Modified and badly glued together 12.27.1999
      by Alessandro De Luca (hairy@pan.medialab.unict.it)
*/

#include "so360private.h"
#include <maverik.h>
#include <sys/time.h>
#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include <math.h>

#define SB_THRESH	-1	/* threshold */
#define SB_MAXVAL	1024	/* 10-bit range */
#define SB_PSCALE	1000	/* position scaling */
#define SB_ASCALE	100	/* angle scaling */
#define SB_PLIMIT	1000000	/* position limit */
#define SB_ALIMIT	18000	/* angle limit */

int SpaceOrbDev=-1;

char *sb_error[] =
{
   "EEPROM storing 0 failed",
   "Receive queue overflow",
   "Transmit queue timeout",
   "Bad packet",
   "Power brown-out",
   "EEPROM checksum error",
   "Hardware fault"
};

unsigned char sb_magic[] = "SpaceWare!";



/* Read a line from the SpaceOrb device */

int SpaceOrbRead(char *buf)
{
  fd_set pf;
  int totread=0;

  do {
    FD_ZERO(&pf);
    FD_SET(SpaceOrbDev, &pf);
    select(200,&pf,NULL,NULL,NULL);
    read(SpaceOrbDev,&buf[totread],1);
  } while (buf[totread++]!=0x0D);

  return totread-1;
}



/* Sync with the SpaceOrb device */

void SpaceOrbSync(void)
{
  char buf[1000];
  fd_set pf;

  FD_ZERO(&pf);
  FD_SET(SpaceOrbDev,&pf);

  /* read until we hit a CR char */
  while (select(200,&pf,NULL,NULL,NULL))
  {
    FD_ZERO(&pf);
    FD_SET(SpaceOrbDev,&pf);
    read(SpaceOrbDev,buf,1);
    if (buf[0]==0x0D) break;
  }
}



/* Open a connection with the SpaceORB device */

int SpaceOrbOpen(char *name)
{
  struct termios mysettings;

  /* Open the serial port to the SpaceORB device */
  SpaceOrbDev=open(name,O_RDWR);

  if (SpaceOrbDev==-1)
  {
    fprintf(stderr,"Could not open SpaceORB device on device %s : ",name);
    perror("");
    return(0);
  }

  /* Read the present state of the port */
  if (tcgetattr(SpaceOrbDev,&mysettings)==-1)
  {
    perror("Could not read SpaceORB ioctl settings ");
    return(0);
  }

  /* Update it with our settings */
  mysettings.c_iflag=IGNBRK|IGNPAR;
  mysettings.c_cc[CEOF]=1;
  mysettings.c_lflag=0;
  cfsetispeed(&mysettings,B9600);
#ifndef TDM_IRIX5
  cfsetospeed(&mysettings,B9600);
#endif

  /* Now set those settings onto the port */
  if (tcsetattr(SpaceOrbDev,TCSANOW,&mysettings)==-1)
  {
    perror("Could not set SpaceORB ioctl settings ");
    return(0);
  }

  /* Configure the SpaceORB device */
  sleep(1);
  SpaceOrbSync();

  return 1;
}



/*
   Convert and scale delta

   NOTE: ideally scales and thresholds should be one per axis
         translation and rotation, to get customizable sensitivity.
*/

void deltaConvert(float *var,int delta,int scale,int threshold)
{
   if (delta&0x200) delta-=1024; /* adjust sign */

   if (threshold!=-1)
   {
      if ((delta>-threshold)||(delta<threshold))
      {
         *var=(float)0.0; /* kill the dwarf!!! */
         return;
      }
   }
   *var=((float)delta)/scale;
}



/*
   Rotate on axis, function taken and modified from TDM keys driver

   NOTE: Maverik functions should be removed from the demon part,
         compilation of this driver should rely only on TDM code.
*/

MAV_quaternion quaternionSet(MAV_vector ax, float ang)
{
  MAV_quaternion q;
  float sa, ca;

/* Convert to radians */

  ang*=MAV_PI_OVER_180;

/* Form the quaternion */

  sa= sin(ang*0.5);
  ca= cos(ang*0.5);

  q.x= ax.x*sa;
  q.y= ax.y*sa;
  q.z= ax.z*sa;
  q.w= ca;

  return q;
}

MAV_matrix matrixQuaternionConvert(MAV_quaternion q)
{
  MAV_matrix m;

  m.mat[0][0]=1.0-2.0*(q.y*q.y+q.z*q.z);
  m.mat[0][1]=2.0*(q.x*q.y-q.w*q.z);
  m.mat[0][2]=2.0*(q.x*q.z+q.w*q.y);
  m.mat[0][3]=0.0;

  m.mat[1][0]=2.0*(q.x*q.y+q.w*q.z);
  m.mat[1][1]=1.0-2.0*(q.x*q.x+q.z*q.z);
  m.mat[1][2]=2.0*(q.y*q.z-q.w*q.x);
  m.mat[1][3]=0.0;

  m.mat[2][0]=2.0*(q.x*q.z-q.w*q.y);
  m.mat[2][1]=2.0*(q.y*q.z+q.w*q.x);
  m.mat[2][2]=1.0-2.0*(q.x*q.x+q.y*q.y);
  m.mat[2][3]=0.0;

  m.mat[3][0]=0.0;
  m.mat[3][1]=0.0;
  m.mat[3][2]=0.0;
  m.mat[3][3]=1.0;

  return m;
}

MAV_vector vectorSet(float x, float y, float z)
{
  MAV_vector rv;
  
  rv.x= x;
  rv.y= y;
  rv.z= z;

  return rv;
}

MAV_vector vectorMult(MAV_vector in, MAV_matrix matrix)
{
  MAV_vector out;

  out.x = in.x*matrix.mat[0][0] + in.y*matrix.mat[0][1] + in.z*matrix.mat[0][2] + matrix.mat[0][3];
  out.y = in.x*matrix.mat[1][0] + in.y*matrix.mat[1][1] + in.z*matrix.mat[1][2] + matrix.mat[1][3];
  out.z = in.x*matrix.mat[2][0] + in.y*matrix.mat[2][1] + in.z*matrix.mat[2][2] + matrix.mat[2][3];
  
  return out;
}

void axisRotate(float *axis,float *v1,float *v2,float amt)
{
  MAV_matrix m;
  MAV_vector nv1,nv2;

  /* Calculate rotation matrix for <amt> degrees about axis */
  m= matrixQuaternionConvert(
     quaternionSet(vectorSet(axis[0],axis[1],axis[2]),amt));

  /* Calculate the resultant vectors */
  nv1=vectorMult(vectorSet(v1[0],v1[1],v1[2]),m);
  nv2=vectorMult(vectorSet(v2[0],v2[1],v2[2]),m);

  /* Update vectors */
  v1[0]=nv1.x; v1[1]=nv1.y; v1[2]=nv1.z;
  v2[0]=nv2.x; v2[1]=nv2.y; v2[2]=nv2.z;
}



/*
   Endlessly read the SpaceORB device

   NOTE: never got any RESET, ERROR, NULLZONE or PULSE packet
         from the device, only ERROR is currently handled.
*/

void SpaceOrbLoop(int debug)
{
   int old_status[TDM_MAX_BUTTONS], firsttime=1, j, k, num, tmp, lc=0;
   char b[1000];
   float xp,yp,zp, xr,yr,zr;
   struct timeval st,ed;

   gettimeofday(&st,NULL);

   while (1)
   {
      lc++;

      num=SpaceOrbRead(b);

      /* Strip the MSB, which is a parity bit */
      for (j=0; j<num; j++) b[j]&=0x7F;

      switch (b[0])
      {
      case 'R':
	if (debug) printf("\nRESET packet (%d/?)\n",num);
	break;
	
      case 'D':
	if (debug) printf("\nDATA packet (%d/12):\n",num);
	if (num!=12) break;

	/* Don't know why Spacetec IMC did this amazing joke... */
	for (j=0; j<9; j++) b[j+2]^=sb_magic[j];

	/* Turn 8.5 chars into six 10-bit integers, then to floats */
	/* axes[0][0-2] are translations, axes[0][3-5] are rotations */
   
	tmp=(((int)b[2]&0x7f)<<3)|(((int)b[3]&0x70)>>4);
	if (debug) printf(" x=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&xp,tmp,SB_PSCALE,SB_THRESH);

	tmp=(((int)b[3]&0x0f)<<6)|(((int)b[4]&0x7e)>>1);
	if (debug) printf(" y=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&yp,tmp,SB_PSCALE,SB_THRESH);

	tmp=(((int)b[4]&0x01)<<9)|(((int)b[5]&0x7f)<<2)|(((int)b[6]&0x60)>>5);
	if (debug) printf(" z=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&zp,tmp,SB_PSCALE,SB_THRESH);

	tmp=(((int)b[6]&0x1f)<<5)|(((int)b[7]&0x7c)>>2);
	if (debug) printf(" a=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&xr,tmp,SB_ASCALE,SB_THRESH);

	tmp=(((int)b[7]&0x03)<<8)|(((int)b[8]&0x7f)<<1)|(((int)b[9]&0x40)>>6);
	if (debug) printf(" b=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&yr,tmp,SB_ASCALE,SB_THRESH);

	tmp=(((int)b[9]&0x3f)<<4)|(((int)b[10]&0x78)>>3);
	if (debug) printf(" c=%+04d",(tmp&0x200)?(tmp-1024):tmp);
	deltaConvert(&zr,tmp,SB_ASCALE,SB_THRESH);

	if (debug) printf("\nxp=%g yp=%g zp=%g xr=%g yr=%g zr=%g\n",xp,yp,zp,xr,yr,zr);

	/* Store the position and local cordinate axis in the shared memory structure. */

	shared->pos[0].xpos+=xp;
	shared->pos[0].ypos+=yp;
	shared->pos[0].zpos+=zp;

	axisRotate(shared->pos[0].u, shared->pos[0].v, shared->pos[0].n,xr);
	axisRotate(shared->pos[0].v, shared->pos[0].u, shared->pos[0].n,yr);
	axisRotate(shared->pos[0].n, shared->pos[0].u, shared->pos[0].v,zr);

	shared->pos[0].status= 0;

	if (debug) printf("\t");
	for (j=0; j<7; j++)
	{
	  if (debug) printf("%s",(b[1]&(1<<j))?"ON  ":"OFF ");
	  shared->new_status[0][j]=(b[1]&(1<<j))?1:0;
	}
	if (debug) printf("\n");

	break;

      case 'K':
	if (debug) printf("\nKEYBT packet (%d/5):\n",num);
	if (num!=5) break;
	if (debug) printf("\t");
	for (j=0; j<7; j++)
	{
	  if (debug) printf("%s",(b[2]&(1<<j))?"ON  ":"OFF ");
	  shared->new_status[0][j]=(b[2]&(1<<j))?1:0;
	}
	if (debug) printf("\n");

	/* 
	   For the first time through the loop case set the old status of the 
	   buttons to the new status so we have something to check against.
	*/

	if (firsttime) {
	  firsttime = 0;
	  for (j=0; j<TDM_MAX_BUTTONS; j++) {
	    old_status[j] = shared->new_status[0][j];
	  }
	}

	/*
	  Loop over all buttons. Check to see if the new status this
	  time around the loop is different to the status last time around the 
	  loop. If it is then the button was either pressed or released, the
	  action is given by the new status.
	*/
    
	for (j=0; j<TDM_MAX_BUTTONS; j++) {	
	  if (shared->new_status[0][j] != old_status[j]) {
	    /* Update event queue - All data is placed in the event_in index of the queue array*/
	  
	    /*
	      Copy the latest position and orientation of each active tracker so as we
	      know this information the (more-or-less) instance the event occurred. 
	    */

	    for (k=0; k<TDM_MAX_TRACKERS; k++) TDM_getPosition(k, &shared->event[shared->event_in].pos[k]);
	    
	    /* 
	       Copy which mouse, button and movement caused the event and the status
	       of all other buttons. 
	    */

	    shared->event[shared->event_in].tracker = 0;
	    shared->event[shared->event_in].button = j;
	    shared->event[shared->event_in].movement = shared->new_status[0][j];
	  
	    TDM_getButtons(shared->event[shared->event_in].button_status);
	  
	    /* Update the pointer to the position where the next event is to be stored */
	    shared->event_in++;
	    if (shared->event_in == MAX_NO_EVENTS) shared->event_in=0;
	  
	    /* 
	       If the new event in position is the same as the event out position then there is no room
	       for the next event. In this case decrease the event in pointer by one. This has the effect
	       that the next event will overwrite the most resent event. This is good since we will always
	       have the last event, and each event carries the state of all buttons.
	    */
	    
	    if (shared->event_in == shared->event_out) shared->event_in--;
	    if (shared->event_in == -1) shared->event_in=MAX_NO_EVENTS-1;
	  }
	
	  /*
	    Update the old status of the mouse to the new status ready for the next
	    time around the loop.
	  */
	  
	  old_status[j] = shared->new_status[0][j];
	}
      
	/* the reset button to restore default pos/angle */
	if (b[2]&0x40)
	{
	  shared->pos[0].xpos=0;
	  shared->pos[0].ypos=0;
	  shared->pos[0].zpos=20; /* really need 20 ??? */
	  shared->pos[0].u[0]=1;
	  shared->pos[0].u[1]=0;
	  shared->pos[0].u[2]=0;
	  shared->pos[0].v[0]=0;
	  shared->pos[0].v[1]=1;
	  shared->pos[0].v[2]=0;
	  shared->pos[0].n[0]=0;
	  shared->pos[0].n[1]=0;
	  shared->pos[0].n[2]=1;
	}
	break;

      case 'E':
	if (debug) printf("\nERROR packet (%d/4):\n",num);
	if (num!=4) break;
	if (debug)
	{
	  for (j=0; j<7; j++)
	    if (b[1]&(1<<j)) printf("\t%s\n",sb_error[j]);
	}
	break;

      case 'N':
	if (debug) printf("\nNZONE packet (%d/3)\n",num);
	break;

      case 'P':
	if (debug) printf("\nPULSE packet (%d/4)\n",num);
	break;

      default:
	if (debug)
	{
	  printf("\nUnknown packet ID %02X (%d/?):\n\t",b[0],num);
	  for (j=0; j<num; j++)
	    printf("%02X ",b[j]);
	  printf("\n");
	}
	break;
      }

      /* Print out record if debugging set */
      if (lc==100&&debug)
      {
	gettimeofday(&ed, NULL);
	printf("time %f\n",lc*1.0/((ed.tv_sec-st.tv_sec)+(ed.tv_usec-st.tv_usec)/1.0E+6));
	printf("[%s]\n",b);
	lc=0;
	st=ed;
      }
   } /* endless loop */
}



/* Routine to close the SpaceORB device */

void SpaceOrbClose(void)
{
  printf("Closing SpaceORB device\n");
  close(SpaceOrbDev);
}
