/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "tdm.h"
#include "maverik.h"
#include <stdio.h>
#include <math.h>

int verbose=0;
TDM_position pos[TDM_MAX_TRACKERS];



/* Routine to initialize a trackers position */

void initTrk(TDM_position *pos)
{
  /* Position */
  pos->xpos=0;
  pos->ypos=0;
  pos->zpos=20;

  /* Trackers axis in transmitter frame */
  pos->u[0]=1; /* X axis */
  pos->u[1]=0;
  pos->u[2]=0;

  pos->v[0]=0; /* Y axis */
  pos->v[1]=1;
  pos->v[2]=0;

  pos->n[0]=0; /* Z axis */
  pos->n[1]=0;
  pos->n[2]=1;

  pos->status=0;
}



/* Routine to initialize the trackers */

int TDM_init(void)
{
  int i;

  if (verbose) printf("initialise trackers\n");

  /* Initialise trackers */
  for (i=0; i<TDM_MAX_TRACKERS; i++) initTrk(&pos[i]);

  /* Return TRUE for successful initialization */
  return 1;
}



/* Routine to rotate vectors v1 and v2 about axis */

void axisRotate(float *axis, float *v1, float *v2)
{
  MAV_matrix m;
  MAV_vector nv1, nv2;

  /* Calculate rotation matrix for 1 degree about axis */
  m= mav_matrixQuaternionConvert(mav_quaternionSet(mav_vectorSet(axis[0], axis[1], axis[2]), 1.0));

  /* Calculate the resultant vectors */
  nv1= mav_vectorMult(mav_vectorSet(v1[0], v1[1], v1[2]), m);
  nv2= mav_vectorMult(mav_vectorSet(v2[0], v2[1], v2[2]), m);

  /* Update vectors */
  v1[0]= nv1.x;
  v1[1]= nv1.y;
  v1[2]= nv1.z;

  v2[0]= nv2.x;
  v2[1]= nv2.y;
  v2[2]= nv2.z;
}



/* Routine to get a single tracker's position */

void TDM_getPosition(int tracker, TDM_position *trk)
{
  if (verbose) printf("get position of tracker %i\n", tracker);

  /* Simulate tracker movement with key presses */
  if (mav_keyboardGet(MAV_KEY_LEFT)==MAV_PRESSED)  /* Move left */
  {
    pos[0].xpos-=0.02;
  } 
  else if (mav_keyboardGet(MAV_KEY_RIGHT)==MAV_PRESSED)  /* Move right */
  {
    pos[0].xpos+=0.02;
  } 
  else if (mav_keyboardGet(MAV_KEY_UP)==MAV_PRESSED)  /* Move up */
  {
    pos[0].ypos+=0.02;
  } 
  else if (mav_keyboardGet(MAV_KEY_DOWN)==MAV_PRESSED)  /* Move down */
  {
    pos[0].ypos-=0.02;
  } 
  else if (mav_keyboardGet('q')==MAV_PRESSED)  /* Move in */
  {
    pos[0].zpos-=0.02;
  } 
  else if (mav_keyboardGet('a')==MAV_PRESSED)  /* Move out */
  {
    pos[0].zpos+=0.02;
  } 
  else if (mav_keyboardGet('x')==MAV_PRESSED)  /* Rotate about x */
  {
    axisRotate(pos[0].u, pos[0].v, pos[0].n);
  } 
  else if (mav_keyboardGet('y')==MAV_PRESSED)  /* Rotate about y */
  {
    axisRotate(pos[0].v, pos[0].u, pos[0].n);
  } 
  else if (mav_keyboardGet('z')==MAV_PRESSED)  /* Rotate about z */
  {
    axisRotate(pos[0].n, pos[0].u, pos[0].v);
  }

  /* Return the tracker's position and orientation */
  *trk= pos[tracker];
}



/* Routine to get the status (pressed/released) of the buttons */

void TDM_getButtons(int button_status[TDM_MAX_TRACKERS][TDM_MAX_BUTTONS])
{
  if (verbose) printf("get buttons\n");
}



/* Routine to get any outstanding button events */

int TDM_getEvent(TDM_buttonEvent *event)
{
  int i, rv=0;

  if (verbose) printf("checking for any events\n");

  /* Simulate event */
  if (mav_keyboardGet('e')==MAV_PRESSED) {

    /* Fill in cause of event */
    event->tracker= 0;
    event->button= 0;
    event->movement= 1; /* 1=pressed, 0=released opposite to Maverik and X */
    
    /* Fill in tracker positions at the time of event */
    for (i=0; i<TDM_MAX_TRACKERS; i++) TDM_getPosition(i, &event->pos[i]);

    /* Fill in button status at the time of event */
    TDM_getButtons(event->button_status);

    /* Return a TRUE value to indicate an event */
    rv=1;
  }

  return rv;
}
