/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "mouseprivate.h"
#include "maverik.h"
#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>

#include <termios.h>
#include <unistd.h>

int mouseBox= -1;



int mouseOpen(char *port)
{
  struct termios tty;

  /* Open the serial port to the mouse */
  mouseBox = open(port, O_RDWR);

  if (mouseBox == -1) {
    fprintf(stderr, "Could not open mouse on device %s : ", port);
    perror("");
    return(0);
  }

  /* Read the present state of the port */
  if (tcgetattr(mouseBox, &tty) == -1) {
    perror("Could not read mouse ioctl settings ");
    return(0);
  }

  /* Update it with our settings */
  tty.c_iflag = IGNBRK | IGNPAR;
  tty.c_oflag = 0;
  tty.c_lflag = 0;
  /*  tty.c_line = 0;*/
  tty.c_cc[VTIME] = 0;
  tty.c_cc[VMIN] = 1;

  cfsetispeed(&tty, B1200);
  cfsetospeed(&tty, B1200);

  /* Now set those settings onto the port */
  if (tcsetattr(mouseBox, TCSAFLUSH, &tty) == -1) {
    perror("Could not set mouse ioctl settings ");
    return(0);
  }

  fprintf(stderr, "Opened connection to serial mouse on %s\n", port);

  return 1;
}



/* Routine to rotate vectors about axis */

MAV_quaternion quaternionSet(MAV_vector ax, float ang)
{
  MAV_quaternion q;
  float sa, ca;

/* Convert to radians */

  ang*=MAV_PI_OVER_180;

/* Form the quaternion */

  sa= sin(ang*0.5);
  ca= cos(ang*0.5);

  q.x= ax.x*sa;
  q.y= ax.y*sa;
  q.z= ax.z*sa;
  q.w= ca;

  return q;
}

MAV_vector vectorSet(float x, float y, float z)
{
  MAV_vector rv;
  
  rv.x= x;
  rv.y= y;
  rv.z= z;

  return rv;
}

MAV_matrix matrixQuaternionConvert(MAV_quaternion q)
{
  MAV_matrix m;

  m.mat[0][0]=1.0-2.0*(q.y*q.y+q.z*q.z);
  m.mat[0][1]=2.0*(q.x*q.y-q.w*q.z);
  m.mat[0][2]=2.0*(q.x*q.z+q.w*q.y);
  m.mat[0][3]=0.0;

  m.mat[1][0]=2.0*(q.x*q.y+q.w*q.z);
  m.mat[1][1]=1.0-2.0*(q.x*q.x+q.z*q.z);
  m.mat[1][2]=2.0*(q.y*q.z-q.w*q.x);
  m.mat[1][3]=0.0;

  m.mat[2][0]=2.0*(q.x*q.z-q.w*q.y);
  m.mat[2][1]=2.0*(q.y*q.z+q.w*q.x);
  m.mat[2][2]=1.0-2.0*(q.x*q.x+q.y*q.y);
  m.mat[2][3]=0.0;

  m.mat[3][0]=0.0;
  m.mat[3][1]=0.0;
  m.mat[3][2]=0.0;
  m.mat[3][3]=1.0;

  return m;
}

MAV_vector vectorMult(MAV_vector in, MAV_matrix matrix)
{
  MAV_vector out;

  out.x = in.x*matrix.mat[0][0] + in.y*matrix.mat[0][1] + in.z*matrix.mat[0][2] + matrix.mat[0][3];
  out.y = in.x*matrix.mat[1][0] + in.y*matrix.mat[1][1] + in.z*matrix.mat[1][2] + matrix.mat[1][3];
  out.z = in.x*matrix.mat[2][0] + in.y*matrix.mat[2][1] + in.z*matrix.mat[2][2] + matrix.mat[2][3];
  
  return out;
}

void axisRotate(float *axis, float am, float *v1, float *v2, float *v3)
{
  MAV_matrix m;
  MAV_vector nv1, nv2, nv3;

  /* Calculate rotation matrix for am degree about axis */
  m= matrixQuaternionConvert(quaternionSet(vectorSet(axis[0], axis[1], axis[2]), am));

  /* Calculate the resultant vectors */
  nv1= vectorMult(vectorSet(v1[0], v1[1], v1[2]), m);
  nv2= vectorMult(vectorSet(v2[0], v2[1], v2[2]), m);
  nv3= vectorMult(vectorSet(v3[0], v3[1], v3[2]), m);

  /* Update vectors */
  v1[0]= nv1.x;
  v1[1]= nv1.y;
  v1[2]= nv1.z;

  v2[0]= nv2.x;
  v2[1]= nv2.y;
  v2[2]= nv2.z;

  v3[0]= nv3.x;
  v3[1]= nv3.y;
  v3[2]= nv3.z;
}



void mouseLoop(int dbg)
{
  int bytes_read, x, y;
  unsigned char c1, c2, c3;
  int lc=0;
  struct timeval st, ed;
  int i, oldbut[2]={0,0};

  gettimeofday(&st, NULL);

  for (;;) {
	
    c1= c2= c3= 0;
	
    do {
      bytes_read= read(mouseBox, &c1, 1);
    } while ((bytes_read == 1) && (c1 & 64) != 64);
      
    if (bytes_read == 1) {
	
      /* want to wait until these bytes are available */
      while (read(mouseBox, &c2, 1) != 1);
      while (read(mouseBox, &c3, 1) != 1);

#if 0
      printf("\nc1 - %i%i%i%i%i%i%i%i\n", (c1&128)/128, (c1&64)/64, (c1&32)/32, (c1&16)/16, (c1&8)/8, (c1&4)/4, (c1&2)/2, (c1&1)/1);
      printf("c2 - %i%i%i%i%i%i%i%i\n", (c2&128)/128, (c2&64)/64, (c2&32)/32, (c2&16)/16, (c2&8)/8, (c2&4)/4, (c2&2)/2, (c2&1)/1);
      printf("c3 - %i%i%i%i%i%i%i%i\n", (c3&128)/128, (c3&64)/64, (c3&32)/32, (c3&16)/16, (c3&8)/8, (c3&4)/4, (c3&2)/2, (c3&1)/1);
#endif

      shared->but[0]=(c1 & 32)/32;
      shared->but[1]=(c1 & 16)/16;
      
      for (i=0; i<2; i++) {
	if (shared->but[i]!=oldbut[i]) {
	  int k;

	  for (k=0; k<TDM_MAX_TRACKERS; k++) TDM_getPosition(k, &shared->event[shared->event_in].pos[k]);

	  shared->event[shared->event_in].tracker = 0;
	  shared->event[shared->event_in].button = i;
	  shared->event[shared->event_in].movement = shared->but[i];

	  TDM_getButtons(shared->event[shared->event_in].button_status);

	  shared->event_in++;
	  if (shared->event_in == MAX_NO_EVENTS) shared->event_in=0;

	  if (shared->event_in == shared->event_out) shared->event_in--;
	  if (shared->event_in == -1) shared->event_in=MAX_NO_EVENTS-1;

	  oldbut[i]=shared->but[i];
	}
      }

      x= ((c1 & 3) << 6) + (c2 & 63);
      y= ((c1 & 12) << 4) + (c3 & 63);
      
      if (x > 127) x -= 256;
      if (y > 127) y -= 256;
      
      shared->x+= x;
      shared->y+= y;
      {
	float yaxis[3]= {0,1,0};
	axisRotate(yaxis, -x*0.1, shared->u, shared->v, shared->n);
	axisRotate(shared->u, y*0.1, shared->u, shared->v, shared->n);
      }

      lc++;
      if (lc==100 && dbg) {
	gettimeofday(&ed, NULL);
	printf("dxy %i %i  cxy %i %i time %f\n", x, y, shared->x, shared->y, lc*1.0/((ed.tv_sec-st.tv_sec)+(ed.tv_usec-st.tv_usec)/1.0E+6));
	lc=0;
	st=ed;
      }
    }
  }
}



void mouseClose(void)
{
  fprintf(stderr, "Closing connection with serial mouse\n");
  close(mouseBox);
}
