/*
 TDM - various "Three D Mice" implementations for GNU Maverik
 Copyright (C) 1999-2001  Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - tdm@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "tdm.h"

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>



typedef int  (*TDM_initFn)(void);
typedef void (*TDM_getPositionFn)(int, TDM_position *);
typedef void (*TDM_getButtonsFn)(int [TDM_MAX_TRACKERS][TDM_MAX_BUTTONS]);
typedef int  (*TDM_getEventFn)(TDM_buttonEvent *);

typedef struct {
  int trk;
  TDM_initFn initFn;
  TDM_getPositionFn getPositionFn;
  TDM_getButtonsFn getButtonsFn;
  TDM_getEventFn getEventFn;
} TDM_device;

TDM_device devs[TDM_MAX_TRACKERS];
int trks[TDM_MAX_TRACKERS];
int numDev=0;



int openlib(int dev, char *libname)
{
  char fn[500];
  void *dlhnd;
  int rv=0;

  /* Get name of library */
  sprintf(fn, "%s/lib/libtdm-%s.so", getenv("TDM_HOME"), libname);

  /* Open library */
#ifdef TDM_SUNOS4
  dlhnd= dlopen(fn, 1);
#else
  dlhnd= dlopen(fn, RTLD_NOW);
#endif  

  if (dlhnd)
  {
    /* Get symbols */
    devs[dev].initFn= (TDM_initFn) dlsym(dlhnd, "TDM_init");

    if (devs[dev].initFn)
    {
      devs[dev].getPositionFn= (TDM_getPositionFn) dlsym(dlhnd, "TDM_getPosition");

      if (devs[dev].getPositionFn)
      {
        devs[dev].getButtonsFn= (TDM_getButtonsFn) dlsym(dlhnd, "TDM_getButtons");
	  
        if (devs[dev].getButtonsFn)
        {
          devs[dev].getEventFn= (TDM_getEventFn) dlsym(dlhnd, "TDM_getEvent");

          if (devs[dev].getEventFn)
          {
	    /* All OK, execute init fn */
	    rv= (*devs[dev].initFn)();
	    
	    if (rv) 
	    {
	      fprintf(stderr, "OK\n");
	    }
	    else
	    {
	      fprintf(stderr, "failed to initialise\n");
	    }
          }
          else
          {
            fprintf(stderr, "failed - could not find TDM_getEvent function in %s, ignoring\n", fn);
          }
        }
        else
        {
          fprintf(stderr, "failed - could not find TDM_getButtons function in %s, ignoring\n", fn);
        }
      }
      else
      {
        fprintf(stderr, "failed - could not find TDM_getPosition function in %s, ignoring\n", fn);
      }
    }
    else
    {
      fprintf(stderr, "failed - could not find TDM_init function in %s, ignoring\n", fn);
    }
  }
  else
  {    
    fprintf(stderr, "failed - could not dlopen TDM library %s, ignoring\n", fn);
  }

  return rv;
}



int TDM_init(void)
{
  FILE *f;
  char lib[500];
  int rv=0, i, trk;

  for (i=0; i<TDM_MAX_TRACKERS; i++) trks[i]=-1;

  if (!getenv("TDM_HOME")) {
    fprintf(stderr, "Error: TDM_HOME variable not set\n");
    return 0;
  }

  /* Try local .tdmrc config file */
  f= fopen(".tdmrc", "r");

  /* Then try ${HOME}/.tdmrc config file */
  if (!f && getenv("HOME")) {
    char fn[500];
    sprintf(fn, "%s/.tdmrc", getenv("HOME"));
    f= fopen(fn, "r");
  }

  /* Last go, try ${TDM_HOME}/.tdmrc config file */
  if (!f) {
    char fn[500];
    sprintf(fn, "%s/.tdmrc", getenv("TDM_HOME"));
    f= fopen(fn, "r");
  }

  if (!f) {
    fprintf(stderr, "Error: could not find .tdmrc, ${HOME}/.tdmrc or ${TDM_HOME}/.tdmrc TDM config files\n");
    return 0;
  }

  while (!feof(f)) {
    
    /* Read tracker number */
    if (fscanf(f, "%i", &trk)==1) {

      /* Read name of library */
      fscanf(f, "%s", lib);    

      fprintf(stderr, "TDM tracker %i is supplied by %s library... ", trk, lib);

      /* Open this library */
      if (openlib(numDev, lib)) 
      {
	rv|=1;
	trks[trk]= numDev;
	devs[numDev].trk= trk;
	numDev++;
      }
      else
      {
	trks[trk]= -2;
	rv|=0;
      }
    }
  }

  fclose(f);

  return rv;
}



void TDM_getPosition(int tracker, TDM_position *trk)
{
  int i;
  
  /* Find appropriate device for this tracker */
  for (i=tracker; i>=0 && trks[i]==-1; i--);

  if (i==-1)
  {
    fprintf(stderr, "Warning: Cant find TDM device for tracker %i\n", tracker);
  }
  else
  {
    if (trks[i]==-2)
    {
      fprintf(stderr, "Warning: TDM device for tracker %i failed to load\n", tracker);
    }
    else
    {
      (*devs[trks[i]].getPositionFn)(tracker-i, trk);
    }
  }
}



void TDM_getButtons(int button_status[TDM_MAX_TRACKERS][TDM_MAX_BUTTONS])
{
  int i, j, k;

  for (i=0; i<numDev; i++) {

    /* Get button status for each device */
    int dbs[TDM_MAX_TRACKERS][TDM_MAX_BUTTONS];
    (*devs[i].getButtonsFn)(dbs);

    /* Copy these values into the correct location */
    for (j=devs[i].trk; j<TDM_MAX_TRACKERS; j++) {
      for (k=0; k<TDM_MAX_BUTTONS; k++) {
	button_status[j][k]= dbs[j-devs[i].trk][k];
      }
    }
  }
}



int TDM_getEvent(TDM_buttonEvent *event)
{
  int i, rv;

  /* Find first event */
  for (i=0; i<numDev; i++) {
    rv= (*devs[i].getEventFn)(event);
    if (rv) return rv;
  }

  return 0;
}
